@@ -18,6 +18,26 @@
   02111-1307, USA.
 */
 
+#ifndef _ECM_IMPL_H
+#define _ECM_IMPL_H 1
+
+#include "config.h"
+#include "ecm.h"
+
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h> /* needed for size_t */
+#endif
+
+#ifndef TUNE
+#include "ecm-params.h"
+#else
+extern size_t MPZMOD_THRESHOLD;
+extern size_t REDC_THRESHOLD;
+#endif
+extern size_t mpn_mul_lo_threshold[];
+
+#include <stdio.h> /* needed for "FILE *" */
+
 #if  defined (__STDC__)                                 \
   || defined (__cplusplus)                              \
   || defined (_AIX)                                     \
@@ -46,18 +66,25 @@ extern FILE *ECM_STDOUT, *ECM_STDERR;
    parameter with ATTRIBUTE_UNUSED when parameter can't be removed, i.e. 
    for interface consistency reasons */
 #ifdef __GNUC__
+#if    __GNUC__ >= 3
 #define ATTRIBUTE_UNUSED __attribute__ ((unused))
+#else
+#define ATTRIBUTE_UNUSED
+#endif
 #define ATTRIBUTE_CONST __attribute__ ((const))
 #else
 #define ATTRIBUTE_UNUSED
 #define ATTRIBUTE_CONST
 #endif
 
-#ifdef __GNUC__
-#define INLINE inline
-#else
-#define INLINE
-#endif
+/* default B2 choice: pow (B1 * METHOD_COST / 6.0, DEFAULT_B2_EXPONENT) */
+#define DEFAULT_B2_EXPONENT 1.43
+#define PM1_COST 1.0 / 6.0
+#define PP1_COST 2.0 / 6.0
+#define ECM_COST 11.0 / 6.0
+
+/* residues are fully reduced (i.e. in canonical mpz form) */
+#define FULL_REDUCTION
 
 /* if POLYEVALTELLEGEN is defined, use polyeval_tellegen(),
    otherwise use polyeval() */
@@ -71,6 +98,9 @@ extern FILE *ECM_STDOUT, *ECM_STDERR;
 #define TOOM3 3
 #define TOOM4 4
 #define KS 5
+#define NTT 6
+
+#include "sp.h"
 
 /* compile with -DMULT=2 to override default */
 #ifndef MULT
@@ -85,6 +115,10 @@ extern FILE *ECM_STDOUT, *ECM_STDERR;
 #define USE_SHORT_PRODUCT
 #endif
 
+/* If defined, ECM stage 2 uses Montgomery form for computing roots of F,G
+   if S == 1. Very slow right now, as we need an inversion per root */
+/* #define MONT_ROOTS */
+
 /* Use George Woltman's GWNUM library */
 /* Should be defined via -DHAVE_GWNUM by Makefile
 #define HAVE_GWNUM
@@ -103,17 +137,26 @@ extern FILE *ECM_STDOUT, *ECM_STDERR;
 #define ASSERT(expr)   do {} while (0)
 #endif
 
+#ifdef MEMORY_DEBUG
+void tests_free (void *, size_t);
+void tests_memory_set_location (char *, unsigned int);
+#define FREE(ptr,size) tests_free(ptr,size)
+#define MEMORY_TAG tests_memory_set_location(__FILE__,__LINE__)
+#define MEMORY_UNTAG tests_memory_set_location("",0)
+#define MPZ_INIT(x)    {MEMORY_TAG;mpz_init(x);MEMORY_UNTAG;}
+#define MPZ_INIT2(x,n) {MEMORY_TAG;mpz_init2(x,n);MEMORY_UNTAG;}
+#else
+#define FREE(ptr,size) free(ptr)
+#define MEMORY_TAG do{}while(0)
+#define MEMORY_UNTAG do{}while(0)
+#define MPZ_INIT(x) mpz_init(x)
+#define MPZ_INIT2(x,n) mpz_init2(x,n)
+#endif
+
+
+
 /* thresholds */
-#ifndef MUL_KARATSUBA_THRESHOLD
-#define MUL_KARATSUBA_THRESHOLD 32
-#endif
-
-#ifndef DIV_DC_THRESHOLD
-#define DIV_DC_THRESHOLD    (3 * MUL_KARATSUBA_THRESHOLD)
-#endif
-
-#define MPZMOD_THRESHOLD_DEFAULT (3 * DIV_DC_THRESHOLD / 2)
-#define REDC_THRESHOLD_DEFAULT   (2 * DIV_DC_THRESHOLD)
+#define MPN_MUL_LO_THRESHOLD 32
 
 /* base2mod is used when size(2^n+/-1) <= BASE2_THRESHOLD * size(cofactor) */
 #define BASE2_THRESHOLD 1.4
@@ -133,10 +176,8 @@ extern FILE *ECM_STDOUT, *ECM_STDERR;
 #define WANT_FREE_PRIME_TABLE(p) (p < 0.0)
 #define FREE_PRIME_TABLE -1.0
 
-#define MOD_PLAIN 0
-#define MOD_BASE2 1
-#define MOD_MODMULN 2
-#define MOD_REDC 3
+/* 2^n+-1 with n < MOD_MINBASE2 cannot use base-2 reduction */
+#define MOD_MINBASE2 16
 
 /* Various logging levels */
 /* OUTPUT_ALWAYS means print always, regardless of verbose value */
@@ -173,6 +214,18 @@ typedef struct
 } __curve_struct;
 typedef __curve_struct curve;
 
+typedef struct
+{
+  unsigned long d1;
+  unsigned long d2;
+  mpz_t i0;
+  int S;
+} __root_params_t;
+typedef __root_params_t root_params_t;
+
+#define EC_MONTGOMERY_FORM 0
+#define EC_WEIERSTRASS_FORM 1
+
 typedef struct
 {
   unsigned int size_fd; /* How many entries .fd has, always nr * (S+1) */
@@ -183,6 +236,11 @@ typedef struct
   unsigned int rsieve; /* Which residue mod dsieve current .next belongs to */
   int dickson_a;       /* Parameter for Dickson polynomials */
   point *fd;
+#ifdef MONT_ROOTS
+  int form;            /* Montgomery or Weierstrass form */
+  long int i0;         /* i0 as a long int (if it fits) */
+#endif
+  unsigned int size_T; /* How many entries T has */
   mpres_t *T;          /* For temp values. FIXME: should go! */
   curve *X;            /* The curve the points are on */
 } __ecm_roots_state;
@@ -214,7 +272,7 @@ typedef struct
   unsigned int rsieve; /* Which residue mod dsieve current .next belongs to */
   point *fd;           /* for S != 1 */
   mpres_t tmp[4];      /* for S=1 */
-  unsigned int d;      /* Step size for computing roots of G */
+  int dickson_a;       /* Parameter for Dickson polynomials */
 } __pp1_roots_state;
 typedef __pp1_roots_state pp1_roots_state;
 
@@ -261,42 +319,50 @@ double   getprime       (double);
 
 /* pm1.c */
 #define pm1_rootsF __ECM(pm1_rootsF)
-int     pm1_rootsF       (mpz_t, listz_t, unsigned int, unsigned int,
-			  unsigned int, mpres_t *, listz_t, int, mpmod_t);
+int     pm1_rootsF       (mpz_t, listz_t, root_params_t *, unsigned long, 
+                          mpres_t *, listz_t, mpmod_t);
 #define pm1_rootsG_init __ECM(pm1_rootsG_init)
-pm1_roots_state* pm1_rootsG_init  (mpres_t *, mpz_t, unsigned int,
-				   unsigned int, int, mpmod_t);
+pm1_roots_state* pm1_rootsG_init  (mpres_t *, root_params_t *, mpmod_t);
 #define pm1_rootsG __ECM(pm1_rootsG)
-int     pm1_rootsG       (mpz_t, listz_t, unsigned int, pm1_roots_state *, 
+int     pm1_rootsG       (mpz_t, listz_t, unsigned long, pm1_roots_state *, 
                           listz_t, mpmod_t);
 #define pm1_rootsG_clear __ECM(pm1_rootsG_clear)
 void    pm1_rootsG_clear (pm1_roots_state *, mpmod_t);
 
 /* bestd.c */
-#define phi __ECM(phi)
-unsigned long phi (unsigned long);
+#define eulerphi __ECM(eulerphi)
+unsigned long eulerphi (unsigned long);
 #define bestD __ECM(bestD)
-int     bestD (mpz_t, mpz_t, int, unsigned int *, unsigned int *, 
-               unsigned int *, unsigned int *, mpz_t);
+int     bestD (root_params_t *, unsigned long *, unsigned long *, mpz_t, 
+               mpz_t, int, int, double, int, mpmod_t);
 
 /* ecm.c */
 #define choose_S __ECM(choose_S)
-int choose_S (mpz_t);
+int  choose_S (mpz_t);
+#define add3 __ECM(add3)
+void add3 (mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, 
+           mpres_t, mpmod_t, mpres_t, mpres_t, mpres_t);
+void duplicate (mpres_t, mpres_t, mpres_t, mpres_t, mpmod_t, mpres_t, mpres_t,
+                mpres_t, mpres_t);
+
+#define ecm_mul __ECM(ecm_mul)
+void ecm_mul (mpres_t, mpres_t, mpz_t, mpmod_t, mpres_t);
 
 /* ecm2.c */
 #define ecm_rootsF __ECM(ecm_rootsF)
-int     ecm_rootsF       (mpz_t, listz_t, unsigned int, unsigned int,
-			  unsigned int, curve *, int, mpmod_t);
+int     ecm_rootsF       (mpz_t, listz_t, root_params_t *, unsigned long, 
+                          curve *, mpmod_t);
 #define ecm_rootsG_init __ECM(ecm_rootsG_init)
-ecm_roots_state* ecm_rootsG_init (mpz_t, curve *, mpz_t, unsigned int,
-		  unsigned int, unsigned int, unsigned int, int, mpmod_t);
+ecm_roots_state* ecm_rootsG_init (mpz_t, curve *, root_params_t *, 
+                                  unsigned long, unsigned long, mpmod_t);
 #define ecm_rootsG __ECM(ecm_rootsG)
-int     ecm_rootsG       (mpz_t, listz_t, unsigned int, ecm_roots_state *, 
+int     ecm_rootsG       (mpz_t, listz_t, unsigned long, ecm_roots_state *, 
                           mpmod_t);
 #define ecm_rootsG_clear __ECM(ecm_rootsG_clear)
-void    ecm_rootsG_clear (ecm_roots_state *, int, mpmod_t);
-void init_roots_state   (ecm_roots_state *, int, unsigned int, unsigned int, 
-                         double);
+void    ecm_rootsG_clear (ecm_roots_state *, mpmod_t);
+#define ecm_findmatch __ECM(ecm_findmatch)
+long    ecm_findmatch (const unsigned long, root_params_t *, curve *, 
+                       mpmod_t, mpz_t);
 
 /* lucas.c */
 #define pp1_mul_prac __ECM(pp1_mul_prac)
@@ -305,29 +371,37 @@ void  pp1_mul_prac     (mpres_t, unsigned long, mpmod_t, mpres_t, mpres_t,
 
 /* pp1.c */
 #define pp1_rootsF __ECM(pp1_rootsF)
-int   pp1_rootsF       (listz_t, unsigned int, unsigned int, unsigned int,
-                        mpres_t *, listz_t, int, mpmod_t);
+int   pp1_rootsF       (listz_t, root_params_t *, unsigned long, mpres_t *, 
+                        listz_t, mpmod_t);
 #define pp1_rootsG __ECM(pp1_rootsG)
-int   pp1_rootsG   (listz_t, unsigned int, pp1_roots_state *, mpmod_t, mpres_t*);
+int   pp1_rootsG   (listz_t, unsigned long, pp1_roots_state *, mpmod_t, 
+                    mpres_t*);
 #define pp1_rootsG_init __ECM(pp1_rootsG_init)
-pp1_roots_state* pp1_rootsG_init (mpres_t*, mpz_t, unsigned int,
-                                  unsigned int, int, mpmod_t);
+pp1_roots_state* pp1_rootsG_init (mpres_t*, root_params_t *, mpmod_t);
 #define pp1_rootsG_clear __ECM(pp1_rootsG_clear)
 void  pp1_rootsG_clear (pp1_roots_state *, mpmod_t);
 
 /* stage2.c */
 #define stage2 __ECM(stage2)
-int          stage2     (mpz_t, void *, mpmod_t, mpz_t, mpz_t, unsigned int,
-                         int, int, int, char *);
+int          stage2     (mpz_t, void *, mpmod_t, unsigned long, unsigned long,
+                         root_params_t *, int, int, char *, int (*)(void));
 #define init_progression_coeffs __ECM(init_progression_coeffs)
-listz_t init_progression_coeffs (mpz_t, unsigned int, unsigned int,
-                         unsigned int, unsigned int, unsigned int, int);
+listz_t init_progression_coeffs (mpz_t, const unsigned long, const unsigned long, 
+				 const unsigned int, const unsigned int, 
+				 const unsigned int, const int);
+#define init_roots_state __ECM(init_roots_state)
+void init_roots_state   (ecm_roots_state *, const int, const unsigned long, 
+			 const unsigned long, const double);
+#define memory_use __ECM(memory_use)
+double memory_use (unsigned long, unsigned int, unsigned int, mpmod_t);
 
 /* listz.c */
 #define list_mul_mem __ECM(list_mul_mem)
 int          list_mul_mem (unsigned int);
 #define init_list __ECM(init_list)
 listz_t      init_list  (unsigned int);
+#define init_list2 __ECM(init_list2)
+listz_t      init_list2  (unsigned int, unsigned int);
 #define clear_list __ECM(clear_list)
 void         clear_list (listz_t, unsigned int);
 #define list_inp_raw __ECM(list_inp_raw)
@@ -342,6 +416,8 @@ void         list_set   (listz_t, listz_t, unsigned int);
 void         list_revert (listz_t, unsigned int);
 #define list_swap __ECM(list_swap)
 void         list_swap  (listz_t, listz_t, unsigned int);
+#define list_neg __ECM(list_neg)
+void         list_neg (listz_t, listz_t, unsigned int, mpz_t);
 #define list_mod __ECM(list_mod)
 void         list_mod   (listz_t, listz_t, unsigned int, mpz_t);
 #define list_add __ECM(list_add)
@@ -352,8 +428,13 @@ void         list_sub   (listz_t, listz_t, listz_t, unsigned int);
 void         list_mul_z (listz_t, listz_t, mpz_t, unsigned int, mpz_t);
 #define list_gcd __ECM(list_gcd)
 int          list_gcd   (mpz_t, listz_t, unsigned int, mpz_t);
+#define list_mulup __ECM(list_mulup)
+void          list_mulup (mpz_t, listz_t, unsigned int, mpz_t, mpz_t);
 #define list_zero __ECM(list_zero)
 void         list_zero  (listz_t, unsigned int);
+#define list_mul __ECM(list_mul)
+void         list_mul (listz_t, listz_t, unsigned int, int, listz_t,
+    unsigned int, int, listz_t);
 #define list_mul_high __ECM(list_mul_high)
 void      list_mul_high (listz_t, listz_t, listz_t, unsigned int, listz_t);
 #define karatsuba __ECM(karatsuba)
@@ -362,17 +443,37 @@ void         karatsuba  (listz_t, listz_t, listz_t, unsigned int, listz_t);
 void        list_mulmod (listz_t, listz_t, listz_t, listz_t, unsigned int,
                          listz_t, mpz_t);
 #define list_invert __ECM(list_invert)
-int         list_invert (listz_t, listz_t, unsigned int, mpz_t, mpmod_t);
+int         list_invert (listz_t, listz_t, unsigned long, mpz_t, mpmod_t);
 #define PolyFromRoots __ECM(PolyFromRoots)
 void      PolyFromRoots (listz_t, listz_t, unsigned int, listz_t, mpz_t);
 #define PolyFromRoots_Tree __ECM(PolyFromRoots_Tree)
 int       PolyFromRoots_Tree (listz_t, listz_t, unsigned int, listz_t, int, 
                          mpz_t, listz_t*, FILE*, unsigned int);
+
+#define ntt_PolyFromRoots __ECM(ntt_PolyFromRoots)
+void	  ntt_PolyFromRoots (listz_t, listz_t, unsigned long, listz_t,
+		mpzspm_t);
+#define ntt_PolyFromRoots_Tree __ECM(ntt_PolyFromRoots_Tree)
+int       ntt_PolyFromRoots_Tree (listz_t, listz_t, unsigned long, listz_t,
+                         int, mpzspm_t, listz_t *, FILE *);
+#define ntt_polyevalT __ECM(ntt_polyevalT)
+int  ntt_polyevalT (mpzv_t, spv_size_t, mpzv_t *, mpzv_t, mpzspv_t,
+		mpzspm_t, char *);
+#define ntt_mul __ECM(ntt_mul)
+void  ntt_mul (mpzv_t, mpzv_t, mpzv_t, spv_size_t, mpzv_t, int, mpzspm_t);
+#define ntt_PrerevertDivision __ECM(ntt_PrerevertDivision)
+void  ntt_PrerevertDivision (listz_t, listz_t, listz_t, mpzspv_t, mpzspv_t,
+		unsigned long, listz_t, mpzspm_t);
+#define ntt_PolyInvert __ECM(ntt_PolyInvert)
+void	     ntt_PolyInvert (listz_t, listz_t, unsigned long, listz_t,
+		mpzspm_t);
+
 #define PrerevertDivision __ECM(PrerevertDivision)
 int   PrerevertDivision (listz_t, listz_t, listz_t, unsigned int, listz_t,
 			 mpz_t);
 #define PolyInvert __ECM(PolyInvert)
 void         PolyInvert (listz_t, listz_t, unsigned int, listz_t, mpz_t);
+
 #define RecursiveDivision __ECM(RecursiveDivision)
 void  RecursiveDivision (listz_t, listz_t, listz_t, unsigned int,
                          listz_t, mpz_t, int);
@@ -383,6 +484,9 @@ void polyeval (listz_t, unsigned int, listz_t*, listz_t, mpz_t, unsigned int);
 #define polyeval_tellegen __ECM(polyeval_tellegen)
 int polyeval_tellegen (listz_t, unsigned int, listz_t*, listz_t,
 		       unsigned int, listz_t, mpz_t, char *);
+#define TUpTree __ECM(TUpTree)
+void TUpTree (listz_t, listz_t *, unsigned int, listz_t, int, unsigned int,
+		mpz_t, FILE *);
 
 /* toomcook.c */
 #define toomcook3 __ECM(toomcook3)
@@ -417,8 +521,12 @@ void mpmod_init_MODMULN (mpmod_t, mpz_t);
 void mpmod_init_REDC (mpmod_t, mpz_t);
 #define mpmod_clear __ECM(mpmod_clear)
 void mpmod_clear (mpmod_t);
+#define mpmod_pausegw __ECM(mpmod_pausegw)
+void mpmod_pausegw (mpmod_t modulus);
+#define mpmod_contgw __ECM(mpmod_contgw)
+void mpmod_contgw (mpmod_t modulus);
 #define mpres_pow __ECM(mpres_pow)
-void mpres_pow (mpres_t, mpres_t, mpres_t, mpmod_t);
+void mpres_pow (mpres_t, mpres_t, mpz_t, mpmod_t);
 #define mpres_ui_pow __ECM(mpres_ui_pow)
 void mpres_ui_pow (mpres_t, unsigned int, mpres_t, mpmod_t);
 #define mpres_mul __ECM(mpres_mul)
@@ -462,7 +570,9 @@ int  mpres_is_zero (mpres_t, mpmod_t);
 /* mul_lo.c */
 #define ecm_mul_lo_n __ECM(ecm_mul_lo_n)
 void ecm_mul_lo_n (mp_ptr, mp_srcptr, mp_srcptr, mp_size_t);
-
+#define ecm_mul_lo_basecase __ECM(ecm_mul_lo_basecase)
+void ecm_mul_lo_basecase (mp_ptr, mp_srcptr, mp_srcptr, mp_size_t);
+	
 /* median.c */
 #define TMulGen __ECM(TMulGen)
 unsigned int
@@ -481,6 +591,8 @@ unsigned int F_mul (mpz_t *, mpz_t *, mpz_t *, unsigned int, int,
 #define F_mul_trans __ECM(F_mul_trans)
 unsigned int F_mul_trans (mpz_t *, mpz_t *, mpz_t *, unsigned int,
                           unsigned int, mpz_t *);
+#define F_clear __ECM(F_clear)
+void F_clear ();
 
 /* rho.c */
 #define rhoinit __ECM(rhoinit)
@@ -496,11 +608,11 @@ void         mpz_sub_si (mpz_t, mpz_t, int);
 #define mpz_divby3_1op __ECM(mpz_divby3_1op)
 void         mpz_divby3_1op (mpz_t);
 #define ceil_log2 __ECM(ceil_log2)
-unsigned int ceil_log2  (unsigned int);
+unsigned int ceil_log2  (unsigned long);
 #define cputime __ECM(cputime)
-unsigned int cputime    (void);
+long         cputime    (void);
 #define elltime __ECM(elltime)
-unsigned int elltime    (unsigned int, unsigned int);
+long         elltime    (long, long);
 #define test_verbose __ECM(test_verbose)
 int          test_verbose (int);
 #define get_verbose __ECM(get_verbose)
@@ -525,6 +637,12 @@ unsigned int get_random_ui (void);
 void Fgwinit (int);
 void Fgwclear (void);
 void Fgwmul (mpz_t, mpz_t, mpz_t);
+int  gw_ecm_stage1 (mpz_t, curve *, mpmod_t, double, double *, mpz_t);
+#endif
+
+/* redc.asm */
+#ifdef NATIVE_REDC
+void ecm_redc3 (mp_ptr, mp_srcptr, mp_size_t, mp_limb_t);
 #endif
 
 
@@ -552,3 +670,5 @@ void Fgwmul (mpz_t, mpz_t, mpz_t);
    mpz_mul (a, b, t);
 #endif
 #endif
+
+#endif /* _ECM_IMPL_H */
