@@ -22,11 +22,14 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <math.h>
-#include <limits.h> /* for ULONG_MAX */
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-impl.h"
+#include <math.h>
+
+#if HAVE_LIMITS_H
+# include <limits.h>
+#else
+# define ULONG_MAX __GMP_ULONG_MAX
+#endif
 
 /******************************************************************************
 *                                                                             *
@@ -39,18 +42,24 @@
 /* Computes curve parameter A and a starting point (x:1) from a given 
    sigma value.
    If a factor of n was found during the process, returns 1 (and factor 
-   in A), 0 otherwise.
+   in f), 0 otherwise.
 */
 static int
 get_curve_from_sigma (mpz_t f, mpres_t A, mpres_t x, mpz_t sigma, mpmod_t n)
 {
   mpres_t t, u, v, b, z;
   
+  MEMORY_TAG;
   mpres_init (t, n);
+  MEMORY_TAG;
   mpres_init (u, n);
+  MEMORY_TAG;
   mpres_init (v, n);
+  MEMORY_TAG;
   mpres_init (b, n);
+  MEMORY_TAG;
   mpres_init (z, n);
+  MEMORY_UNTAG;
 
   mpres_set_z  (u, sigma, n);
   mpres_mul_ui (v, u, 4, n);   /* v = (4*sigma) mod n */
@@ -109,7 +118,9 @@ montgomery_to_weierstrass (mpz_t f, mpres_t x, mpres_t y, mpres_t A, mpmod_t n)
 {
   mpres_t g;
   
+  MEMORY_TAG;
   mpres_init (g, n);
+  MEMORY_UNTAG;
   mpres_add (g, x, A, n);
   mpres_mul (g, g, x, n);
   mpres_add_ui (g, g, 1, n);
@@ -151,7 +162,7 @@ montgomery_to_weierstrass (mpz_t f, mpres_t x, mpres_t y, mpres_t A, mpmod_t n)
    Modifies: x3, z3, u, v, w.
    (x3,z3) may be identical to (x2,z2) and to (x,z)
 */
-static void
+void
 add3 (mpres_t x3, mpres_t z3, mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, 
       mpres_t x, mpres_t z, mpmod_t n, mpres_t u, mpres_t v, mpres_t w)
 {
@@ -193,7 +204,7 @@ add3 (mpres_t x3, mpres_t z3, mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1,
      - b : (a+2)/4 mod n
      - t, u, v, w : auxiliary variables
 */
-static void
+void
 duplicate (mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, mpmod_t n, 
            mpres_t b, mpres_t u, mpres_t v, mpres_t w)
 {
@@ -208,25 +219,48 @@ duplicate (mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, mpmod_t n,
   mpres_mul (z2, w, u, n);  /* z2 = ((4*x1*z1)*((x1-z1)^2+(A+2)/4*(4*x1*z1))) mod n */
 }
 
-/* multiply P=(x:z) by e and puts the result in (x:z).
-   Assumes e >= 1.
-*/
-static void
+/* multiply P=(x:z) by e and puts the result in (x:z). */
+void
 ecm_mul (mpres_t x, mpres_t z, mpz_t e, mpmod_t n, mpres_t b)
 {
   size_t l;
+  int negated = 0;
   mpres_t x0, z0, x1, z1, u, v, w;
 
-  if (mpz_cmp_ui (e, 1) == 0)
-    return;
+  /* In Montgomery coordinates, the point at infinity is (0::0) */
+  if (mpz_sgn (e) == 0)
+    {
+      mpz_set_ui (x, 0);
+      mpz_set_ui (z, 0);
+      return;
+    }
 
+  /* The negative of a point (x:y:z) is (x:-y:u). Since we do not compute
+     y, e*(x::z) == (-e)*(x::z). */
+  if (mpz_sgn (e) < 0)
+    {
+      negated = 1;
+      mpz_neg (e, e);
+    }
+
+  if (mpz_cmp_ui (e, 1) == 0)
+    goto ecm_mul_end;
+
+  MEMORY_TAG;
   mpres_init (x0, n);
+  MEMORY_TAG;
   mpres_init (z0, n);
+  MEMORY_TAG;
   mpres_init (x1, n);
+  MEMORY_TAG;
   mpres_init (z1, n);
+  MEMORY_TAG;
   mpres_init (u, n);
+  MEMORY_TAG;
   mpres_init (v, n);
+  MEMORY_TAG;
   mpres_init (w, n);
+  MEMORY_UNTAG;
 
   l = mpz_sizeinbase (e, 2) - 1; /* l >= 1 */
 
@@ -260,13 +294,17 @@ ecm_mul (mpres_t x, mpres_t z, mpz_t e, mpmod_t n, mpres_t b)
   mpres_clear (u, n);
   mpres_clear (v, n);
   mpres_clear (w, n);
+
+ecm_mul_end:
+
+  /* Undo negation to avoid changing the caller's e value */
+  if (negated)
+    mpz_neg (e, e);
 }
 
 #define ADD 6.0 /* number of multiplications in an addition */
 #define DUP 5.0 /* number of multiplications in a duplicate */
 
-#define START(d,v) ((d)/1.6180339887498948482-128.0+(v))
-
 /* returns the number of modular multiplications for computing
    V_n from V_r * V_{n-r} - V_{n-2r}.
    ADD is the cost of an addition
@@ -347,7 +385,17 @@ lucas_cost (unsigned long n, double v)
 }
 
 
-/* computes kP from P=(xA:zA) and puts the result in (xA:zA). Assumes k>2. */
+/* computes kP from P=(xA:zA) and puts the result in (xA:zA). Assumes k>2. 
+   WARNING! The calls to add3() assume that the two input points are distinct,
+   which is not neccessarily satisfied. The result can be that in rare cases
+   the point at infinity (z==0) results when it shouldn't. A test case is 
+   echo 33554520197234177 | ./ecm -sigma 2046841451 373 1
+   which finds the prime even though it shouldn't (23^2=529 divides order).
+   This is not a problem for ECM since at worst we'll find a factor we 
+   shouldn't have found. For other purposes (i.e. primality proving) this 
+   would have to be fixed first.
+*/
+
 static void
 prac (mpres_t xA, mpres_t zA, unsigned long k, mpmod_t n, mpres_t b,
       mpres_t u, mpres_t v, mpres_t w, mpres_t xB, mpres_t zB, mpres_t xC, 
@@ -358,9 +406,10 @@ prac (mpres_t xA, mpres_t zA, unsigned long k, mpmod_t n, mpres_t b,
   __mpz_struct *tmp;
 #define NV 10  
   static double val[NV] =
-    { 1.61803398875, 1.72360679775, 1.618347119656, 1.617914406529,
-      1.612429949509, 1.632839806089, 1.620181980807, 1.580178728295,
-      1.617214616534, 1.38196601125 };
+    { 1.6180339887498948, 1.7236067977499790, 1.6183471196562281,
+      1.6179144065288179, 1.6124299495094950, 1.6328398060887063,
+      1.6201819808074158, 1.5801787282954641, 1.6172146165344039,
+      1.3819660112501052 };
   
   /* chooses the best value of v */
   for (d = 0, cmin = ADD * (double) k; d < NV; d++)
@@ -498,31 +547,49 @@ prac (mpres_t xA, mpres_t zA, unsigned long k, mpmod_t n, mpres_t b,
    Return value: non-zero iff a factor is found
 */
 static int
-ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, double B1done,
-	    mpz_t go)
+ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, 
+            double *B1done, mpz_t go, int (*stop_asap)(void))
 {
   mpres_t b, z, u, v, w, xB, zB, xC, zC, xT, zT, xT2, zT2;
   double q, r;
   int ret = 0;
 
+  MEMORY_TAG;
   mpres_init (b, n);
+  MEMORY_TAG;
   mpres_init (z, n);
+  MEMORY_TAG;
   mpres_init (u, n);
+  MEMORY_TAG;
   mpres_init (v, n);
+  MEMORY_TAG;
   mpres_init (w, n);
+  MEMORY_TAG;
   mpres_init (xB, n);
+  MEMORY_TAG;
   mpres_init (zB, n);
+  MEMORY_TAG;
   mpres_init (xC, n);
+  MEMORY_TAG;
   mpres_init (zC, n);
+  MEMORY_TAG;
   mpres_init (xT, n);
+  MEMORY_TAG;
   mpres_init (zT, n);
+  MEMORY_TAG;
   mpres_init (xT2, n);
+  MEMORY_TAG;
   mpres_init (zT2, n);
+  MEMORY_UNTAG;
 
   mpres_set_ui (z, 1, n);
-  
+
   mpres_add_ui (b, A, 2, n);
-  mpres_div_2exp (b, b, 2, n); /* b == (A+2)/4 */
+  mpres_div_2exp (b, b, 2, n); /* b == (A0+2)*B/4, where B=2^(k*GMP_NUMB_LIMB)
+                                  for MODMULN or REDC, B=1 otherwise */
+#ifndef FULL_REDUCTION
+  mpres_semi_normalize (b, mpz_size (n->orig_modulus));
+#endif
 
   /* preload group order */
   if (go != NULL)
@@ -530,12 +597,12 @@ ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, double B1done,
 
   /* prac() wants multiplicands > 2 */
   for (r = 2.0; r <= B1; r *= 2.0)
-    if (r > B1done)
+    if (r > *B1done)
       duplicate (x, z, x, z, n, b, u, v, w);
   
   /* We'll do 3 manually, too (that's what ecm4 did..) */
   for (r = 3.0; r <= B1; r *= 3.0)
-    if (r > B1done)
+    if (r > *B1done)
       {
         duplicate (xB, zB, x, z, n, b, u, v, w);
         add3 (x, z, x, z, xB, zB, x, z, n, u, v, w);
@@ -545,13 +612,31 @@ ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, double B1done,
   for (q = getprime (q); q <= B1; q = getprime (q))
     {
       for (r = q; r <= B1; r *= q)
-	if (r > B1done)
+	if (r > *B1done)
 	  prac (x, z, (unsigned long) q, n, b, u, v, w, xB, zB, xC, zC, xT,
 		zT, xT2, zT2);
+
+      if (mpres_is_zero (z, n))
+        {
+          outputf (OUTPUT_VERBOSE, "Reached point at infinity, %.0f divides "
+                   "group order\n", q);
+          break;
+        }
+
+      if (stop_asap != NULL && (*stop_asap) ())
+        {
+          outputf (OUTPUT_NORMAL, "Interrupted at prime %.0f\n", q);
+          break;
+        }
     }
+  
   getprime (FREE_PRIME_TABLE); /* free the prime tables, and reinitialize */
+  *B1done = (q < B1) ? q : B1;
 
   /* Normalize z to 1 */
+#ifndef FULL_REDUCTION
+  mpres_normalize (z); /* needed for gcd */
+#endif
   if (!mpres_invert (u, z, n)) /* Factor found? */
     {
       mpres_gcd (f, z, n);
@@ -594,6 +679,83 @@ choose_S (mpz_t B2len)
     return -30; /* Dickson(30) */
 }
 
+static void
+print_expcurves (mpz_t B2min, mpz_t effB2, unsigned long dF, unsigned long k, 
+                 int S, int clear)
+{
+  double prob;
+  int i;
+  char sep;
+
+  if (!test_verbose (OUTPUT_VERBOSE))
+    return;
+
+  rhoinit (256, 10);
+  outputf (OUTPUT_VERBOSE, "Expected number of curves to find a factor "
+           "of n digits:\n20\t25\t30\t35\t40\t45\t50\t55\t60\t65\n");
+  for (i = 20; i <= 65; i += 5)
+    {
+      sep = (i < 65) ? '\t' : '\n';
+      prob = ecmprob (mpz_get_d (B2min), mpz_get_d (effB2),
+                      pow (10., i - .5), (double) dF * dF * k, S);
+      if (prob > 1. / 10000000)
+        outputf (OUTPUT_VERBOSE, "%.0f%c", floor (1. / prob + .5), sep);
+      else if (prob > 0.)
+        outputf (OUTPUT_VERBOSE, "%.2g%c", floor (1. / prob + .5), sep);
+      else
+        outputf (OUTPUT_VERBOSE, "Inf%c", sep);
+    }
+
+  if (clear)
+    rhoinit (1, 0); /* Free memory of rhotable */
+}
+
+static void
+print_exptime (mpz_t B2min, mpz_t effB2, unsigned long dF, unsigned long k, 
+               int S, double tottime, int clear)
+{
+  double prob, exptime;
+  int i;
+  char sep;
+  
+  if (!test_verbose (OUTPUT_VERBOSE))
+    return;
+  
+  rhoinit (256, 10);
+  outputf (OUTPUT_VERBOSE, "Expected time to find a factor of n digits:\n"
+    "20\t25\t30\t35\t40\t45\t50\t55\t60\t65\n");
+  for (i = 20; i <= 65; i += 5)
+    {
+      sep = (i < 65) ? '\t' : '\n';
+      prob = ecmprob (mpz_get_d (B2min), mpz_get_d (effB2), 
+                      pow (10., i - .5), (double) dF * dF * k, S);
+      exptime = (prob > 0.) ? tottime / prob : HUGE_VAL;
+      outputf (OUTPUT_TRACE, "Digits: %d, Total time: %.0f, probability: "
+               "%g, expected time: %.0f\n", i, tottime, prob, exptime);
+      if (exptime < 1000.)
+        outputf (OUTPUT_VERBOSE, "%.0fms%c", exptime, sep);
+      else if (exptime < 60000.) /* One minute */
+        outputf (OUTPUT_VERBOSE, "%.2fs%c", exptime / 1000., sep);
+      else if (exptime < 3600000.) /* One hour */
+        outputf (OUTPUT_VERBOSE, "%.2fm%c", exptime / 60000., sep);
+      else if (exptime < 86400000.) /* One day */
+        outputf (OUTPUT_VERBOSE, "%.2fh%c", exptime / 3600000., sep);
+      else if (exptime < 31536000000.) /* One year */
+        outputf (OUTPUT_VERBOSE, "%.2fd%c", exptime / 86400000., sep);
+      else if (exptime < 31536000000000.) /* One thousand years */
+        outputf (OUTPUT_VERBOSE, "%.2fy%c", exptime / 31536000000., sep);
+      else if (exptime < 31536000000000000.) /* One million years */
+        outputf (OUTPUT_VERBOSE, "%.0fy%c", exptime / 31536000000., sep);
+      else if (prob > 0.)
+        outputf (OUTPUT_VERBOSE, "%.1gy%c", exptime / 31536000000., sep);
+      else 
+        outputf (OUTPUT_VERBOSE, "Inf%c", sep);
+    }
+
+  if (clear)
+    rhoinit (1, 0); /* Free memory of rhotable */
+}
+
 /* Input: x is starting point or zero
           sigma is sigma value (if x is set to zero) or 
             A parameter (if x is non-zero) of curve
@@ -613,16 +775,22 @@ choose_S (mpz_t B2len)
 		 ECM_ERROR in case of error.
 */
 int
-ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
+ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double *B1done,
      double B1, mpz_t B2min_parm, mpz_t B2_parm, double B2scale, 
-     unsigned int k, int S, int verbose, int repr, int sigma_is_A, 
-     FILE *os, FILE* es, char *TreeFilename)
+     unsigned long k, const int S, int verbose, int repr, int use_ntt,
+     int sigma_is_A, FILE *os, FILE* es, char *TreeFilename, double maxmem,
+     double stage1time, gmp_randstate_t rng, int (*stop_asap)(void))
 {
   int youpi = ECM_NO_FACTOR_FOUND;
-  unsigned int st;
+  int base2 = 0;  /* If n is of form 2^n[+-]1, set base to [+-]n */
+  int Fermat = 0; /* If base2 > 0 is a power of 2, set Fermat to base2 */
+  int po2 = 0;    /* Whether we should use power-of-2 poly degree */
+  long st;
   mpmod_t modulus;
   curve P;
   mpz_t B2min, B2; /* Local B2, B2min to avoid changing caller's values */
+  unsigned long dF;
+  root_params_t root_params;
 
   set_verbose (verbose);
   ECM_STDOUT = (os == NULL) ? stdout : os;
@@ -637,10 +805,64 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
 
   /* now n is odd */
 
+  /* check that B1 is not too large */
+  if (B1 > (double) ULONG_MAX)
+    {
+      outputf (OUTPUT_ERROR, "Error, maximal step 1 bound for ECM is %lu.\n", 
+               ULONG_MAX);
+      return ECM_ERROR;
+    }
+
   st = cputime ();
 
+  /* See what kind of number we have as that may influence optimal parameter 
+     selection. Test for base 2 number */
+
+  if (repr != ECM_MOD_NOBASE2)
+    base2 = (abs (repr) >= 16) ? repr : isbase2 (n, BASE2_THRESHOLD);
+
+  /* For for a Fermat number (base2 a positive power of 2) */
+  for (Fermat = base2; Fermat > 0 && (Fermat & 1) == 0; Fermat >>= 1);
+  if (Fermat == 1) 
+    {
+      Fermat = base2;
+      po2 = 1;
+    }
+  else
+      Fermat = 0;
+
+  if (base2)
+    {
+      if (mpmod_init_BASE2 (modulus, base2, n) == ECM_ERROR)
+        return ECM_ERROR;
+    }
+  else if (repr == ECM_MOD_MPZ)
+    mpmod_init_MPZ (modulus, n);
+  else if (repr == ECM_MOD_MODMULN)
+    mpmod_init_MODMULN (modulus, n);
+  else if (repr == ECM_MOD_REDC)
+    mpmod_init_REDC (modulus, n);
+  else /* automatic choice of general reduction method */
+    mpmod_init (modulus, n, -1);
+
+  MEMORY_TAG;
+  mpres_init (P.x, modulus);
+  MEMORY_TAG;
+  mpres_init (P.y, modulus);
+  MEMORY_TAG;
+  mpres_init (P.A, modulus);
+
+  mpres_set_z (P.x, x, modulus);
+  mpres_set_ui (P.y, 1, modulus);
+  
+  MEMORY_TAG;
   mpz_init_set (B2min, B2min_parm);
+  MEMORY_TAG;
   mpz_init_set (B2, B2_parm);
+  
+  MEMORY_TAG;
+  mpz_init (root_params.i0);
+  MEMORY_UNTAG;
 
   /* set second stage bound B2: when using polynomial multiplication of
      complexity n^alpha, stage 2 has complexity about B2^(alpha/2), and
@@ -657,25 +879,24 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
   /* Also scale B2 by what the user said (or by the default scaling of 1.0) */
 
   if (ECM_IS_DEFAULT_B2(B2))
-    mpz_set_d (B2, B2scale * pow (11.0 / 6.0 * B1, 1.424828748));
+    mpz_set_d (B2, B2scale * pow (ECM_COST * B1, DEFAULT_B2_EXPONENT));
 
   /* set B2min */
   if (mpz_sgn (B2min) < 0)
     mpz_set_d (B2min, B1);
 
-  if (repr == 1)
-    mpmod_init_MPZ (modulus, n);
-  else if (repr == MOD_MODMULN)
-    mpmod_init_MODMULN (modulus, n);
-  else if (repr == MOD_REDC)
-    mpmod_init_REDC (modulus, n);
-  else if (abs (repr) > 16)
+  /* Let bestD determine parameters for root generation and the 
+     effective B2 */
+
+  if (use_ntt)
+    po2 = 1;
+
+  if (bestD (&root_params, &k, &dF, B2min, B2, po2, use_ntt, maxmem, 
+             (TreeFilename != NULL), modulus) == ECM_ERROR)
     {
-      if (mpmod_init_BASE2 (modulus, repr, n) == ECM_ERROR)
-        return ECM_ERROR;
+      youpi = ECM_ERROR;
+      goto end_of_ecm;
     }
-  else /* automatic choice, avoiding base2 if repr=-1 */
-    mpmod_init (modulus, n, repr);
 
   /* Set default degree for Brent-Suyama extension */
   /* We try to keep the time used by the Brent-Suyama extension
@@ -684,61 +905,48 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
      the better Dickson polys whenever possible. For S == 1, 2, they behave
      identically. */
 
-  if (S == ECM_DEFAULT_S)
+  root_params.S = S;
+  if (root_params.S == ECM_DEFAULT_S)
     {
-      /* This requires that modulus is already inited */
-      if (modulus->repr == MOD_BASE2 && modulus->Fermat > 0)
+      if (Fermat > 0)
         {
           /* For Fermat numbers, default is 1 (no Brent-Suyama) */
-          S = 1;
+          root_params.S = 1;
         }
       else
         {
           mpz_t t;
+          MEMORY_TAG;
           mpz_init (t);
+          MEMORY_UNTAG;
           mpz_sub (t, B2, B2min);
-          S = choose_S (t);
+          root_params.S = choose_S (t);
           mpz_clear (t);
         }
     }
   
-  mpres_init (P.x, modulus);
-  mpres_init (P.A, modulus);
-
   if (sigma_is_A == 0)
     {
       /* if sigma=0, generate it at random */
-      if (mpz_cmp_ui (sigma, 0) == 0)
+      if (mpz_sgn (sigma) == 0)
         {
-          gmp_randstate_t state;
-          gmp_randinit_default (state);
-          gmp_randseed_ui (state, get_random_ui ());
-          mpz_urandomb (sigma, state, 32);
+          mpz_urandomb (sigma, rng, 32);
           mpz_add_ui (sigma, sigma, 6);
-          gmp_randclear (state);
         }
 
-      /* sigma contains sigma value, A value must be computed */
-      outputf (OUTPUT_NORMAL, "Using B1=%1.0f, B2=", B1);
-      if (mpz_cmp_d (B2min, B1) == 0)
-        outputf (OUTPUT_NORMAL, "%Zd", B2);
-      else
-        outputf (OUTPUT_NORMAL, "%Zd-%Zd", B2min, B2);
-      outputf (OUTPUT_NORMAL, ", polynomial ");
-      if (S > 0)
-        outputf (OUTPUT_NORMAL, "x^%u", S);
-      else
-        outputf (OUTPUT_NORMAL, "Dickson(%u)", -S);
-      outputf (OUTPUT_NORMAL, ", sigma=%Zd\n", sigma);
-
-      if ((youpi = get_curve_from_sigma (f, P.A, P.x, sigma, modulus)))
-        goto end_of_ecm;
+      /* sigma contains sigma value, A and x values must be computed */
+      /* If x is specified by user, keep that value */
+      youpi = get_curve_from_sigma (f, P.A, P.x, sigma, modulus);
+      if (youpi != ECM_NO_FACTOR_FOUND)
+	  goto end_of_ecm;
     }
   else
     {
       /* sigma contains the A value */
       mpres_set_z (P.A, sigma, modulus);
       /* TODO: make a valid, random starting point in case none was given */
+      /* Problem: this may be as hard as factoring as we'd need to determine
+         whether x^3 + a*x^2 + x is a quadratic residue or not */
       /* For now, we'll just chicken out. */
       if (mpz_sgn (x) == 0)
         {
@@ -752,12 +960,36 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
   /* If a nonzero value is given in x, then we use it as the starting point */
   if (mpz_sgn (x) != 0)
       mpres_set_z (P.x, x, modulus);
-  
+
+  /* Now that the parameters are decided, print info */
+
+  outputf (OUTPUT_NORMAL, "Using B1=%1.0f, B2=", B1);
+  if (mpz_cmp_d (B2min, B1) == 0)
+      outputf (OUTPUT_NORMAL, "%Zd", B2);
+  else
+      outputf (OUTPUT_NORMAL, "%Zd-%Zd", B2min, B2);
+  outputf (OUTPUT_NORMAL, ", polynomial ");
+  if (root_params.S > 0)
+      outputf (OUTPUT_NORMAL, "x^%u", root_params.S);
+  else
+      outputf (OUTPUT_NORMAL, "Dickson(%u)", -root_params.S);
+  outputf (OUTPUT_NORMAL, ", sigma=%Zd\n", sigma);
+
+#if 0
+  outputf (OUTPUT_VERBOSE, "b2=%1.0f, dF=%lu, k=%lu, d=%lu, d2=%lu, i0=%Zd\n", 
+           b2, dF, k, root_params.d1, root_params.d2, root_params.i0);
+#else
+  outputf (OUTPUT_VERBOSE, "dF=%lu, k=%lu, d=%lu, d2=%lu, i0=%Zd\n", 
+           dF, k, root_params.d1, root_params.d2, root_params.i0);
+#endif
+
   if (test_verbose (OUTPUT_RESVERBOSE))
     {
       mpz_t t;
-      
+
+      MEMORY_TAG;
       mpz_init (t);
+      MEMORY_UNTAG;
       mpres_get_z (t, P.A, modulus);
       outputf (OUTPUT_RESVERBOSE, "a=%Zd\n", t);
       mpres_get_z (t, P.x, modulus);
@@ -768,54 +1000,100 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
   if (go != NULL && mpz_cmp_ui (go, 1) > 0)
     outputf (OUTPUT_VERBOSE, "initial group order: %Zd\n", go);
 
-  /* check that B1 is not too large */
-  if (B1 > (double) ULONG_MAX)
-    {
-      outputf (OUTPUT_ERROR, "Error, maximal step1 bound for ECM is %lu.\n", 
-               ULONG_MAX);
-      youpi = ECM_ERROR;
-      goto end_of_ecm;
-    }
+  print_expcurves (B2min, B2, dF, k, root_params.S, 0);
 
-  if (B1 > B1done)
-    youpi = ecm_stage1 (f, P.x, P.A, modulus, B1, B1done, go);
+#ifdef HAVE_GWNUM
+  /* Right now, we only do base 2 numbers with GWNUM */
+
+  if (base2 != 0 && B1 >= *B1done)
+      youpi = gw_ecm_stage1 (f, &P, modulus, B1, B1done, go);
+
+  /* At this point B1 == *B1done unless interrupted, or no GWNUM ecm_stage1
+     is available */
+
+  if (youpi != ECM_NO_FACTOR_FOUND)
+    goto end_of_ecm;
+
+#endif
+
+  if (B1 > *B1done)
+    youpi = ecm_stage1 (f, P.x, P.A, modulus, B1, B1done, go, stop_asap);
   
-  st = elltime (st, cputime ());
-  
-  outputf (OUTPUT_NORMAL, "Step 1 took %ums\n", st);
+  if (stage1time > 0.)
+    {
+      const long st2 = elltime (st, cputime ());
+      const long s1t = (long) (stage1time * 1000.);
+      outputf (OUTPUT_NORMAL, 
+               "Step 1 took %ldms (%ld in this run, %ld from previous runs)\n", 
+               st2 + s1t, st2, s1t);
+    }
+  else
+    outputf (OUTPUT_NORMAL, "Step 1 took %ldms\n", elltime (st, cputime ()));
 
   /* Store end-of-stage-1 residue in x in case we write it to a save file, 
      before P.x is converted to Weierstrass form */
   
   mpres_get_z (x, P.x, modulus);
 
-  if (youpi != ECM_NO_FACTOR_FOUND) /* a factor was found */
+  if (youpi != ECM_NO_FACTOR_FOUND)
     goto end_of_ecm;
 
   if (test_verbose (OUTPUT_RESVERBOSE)) 
     {
       mpz_t t;
       
+      MEMORY_TAG;
       mpz_init (t);
+      MEMORY_UNTAG;
       mpres_get_z (t, P.x, modulus);
       outputf (OUTPUT_RESVERBOSE, "x=%Zd\n", t);
       mpz_clear (t);
     }
 
-  mpres_init (P.y, modulus);
+  /* In case of a signal, we'll exit after the residue is printed. If no save
+     file is specified, the user may still resume from the residue */
+  if (stop_asap != NULL && (*stop_asap) ())
+    goto end_of_ecm;
 
+#ifdef MONT_ROOTS
+  /* If we want to use Montgomery form for generating the roots for stage 2, 
+     convert to Weierstrass form only if S != 1 */
+  if (root_params.S != 1)
+    {
+      mpz_t t;
+      outputf (OUTPUT_DEVVERBOSE, "ecm: Converting to Weierstrass form\n");
+      youpi = montgomery_to_weierstrass (f, P.x, P.y, P.A, modulus);
+      if (test_verbose (OUTPUT_TRACE))
+        {
+          MEMORY_TAG;
+          mpz_init (t);
+          MEMORY_UNTAG;
+          mpres_get_z (t, P.x, modulus);
+          outputf (OUTPUT_TRACE, "P = (%Zd, ", t);
+          mpres_get_z (t, P.y, modulus);
+          outputf (OUTPUT_TRACE, "%Zd)\n", t);
+          mpz_clear (t);
+        }
+    }
+#else
   youpi = montgomery_to_weierstrass (f, P.x, P.y, P.A, modulus);
+#endif
 
-  if (youpi == ECM_NO_FACTOR_FOUND)
-    youpi = stage2 (f, &P, modulus, B2min, B2, k, S, ECM_ECM, st, 
-                    TreeFilename);
+  if (youpi == ECM_NO_FACTOR_FOUND && mpz_cmp (B2, B2min) >= 0)
+    youpi = stage2 (f, &P, modulus, dF, k, &root_params, ECM_ECM, 
+                    use_ntt, TreeFilename, stop_asap);
   
-  mpres_clear (P.y, modulus);
+  if (youpi == ECM_NO_FACTOR_FOUND && (stop_asap == NULL || !(*stop_asap)()))
+    print_exptime (B2min, B2, dF, k, root_params.S, 
+                   (long) (stage1time * 1000.) + 
+                   elltime (st, cputime ()), 1);
 
 end_of_ecm:
-  mpres_clear (P.x, modulus);
   mpres_clear (P.A, modulus);
+  mpres_clear (P.y, modulus);
+  mpres_clear (P.x, modulus);
   mpmod_clear (modulus);
+  mpz_clear (root_params.i0);
   mpz_clear (B2);
   mpz_clear (B2min);
 
