@@ -21,14 +21,12 @@
 */
 
 #include <math.h>
-#include <stdio.h>
 #include <stdlib.h>
-#include "gmp.h"
-#include "ecm.h"
+
 #ifdef OUTSIDE_LIBECM
-#include "ecm-ecm.h"
+# include "ecm-ecm.h"
 #else
-#include "ecm-impl.h"
+# include "ecm-impl.h"
 #endif
 
 /* This function returns successive odd primes, starting with 3.
@@ -70,8 +68,17 @@ getprime (double pp)
       moduli = NULL;
       return pp;
     }
-
-  while ((++current < len) && (sieve[current] == 0));
+  
+  /* the following complex block is equivalent to:
+     while ((++current < len) && (sieve[current] == 0));
+     but is faster.
+  */
+  {
+    unsigned char *ptr = sieve + current;
+    unsigned char *end = sieve + len;
+    while ((++ptr < end) && (*ptr == 0));
+    current = ptr - sieve;
+  }
 
   if (current < len) /* most calls will end here */
     return offset + 2.0 * (double) current;
@@ -116,7 +123,7 @@ getprime (double pp)
 	  }
 	else
 	  {
-	    unsigned long int i, p, j, ok;
+	    unsigned int i, p, j, ok;
 
 	    i = nprimes;
 	    nprimes *= 2;
