@@ -20,11 +20,7 @@
   MA 02111-1307, USA.
 */
 
-#include <assert.h>
-#include <stdio.h>
 #include <stdlib.h>
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-impl.h"
 
 #ifdef DEBUG
@@ -35,7 +31,7 @@
 
 #if (MULT == KS)
 #define LIST_MULT_N kronecker_schonhage
-#ifdef HAVE_FFT
+#ifdef HAVE___GMPN_MUL_FFT
 #define WRAP /* use wrap-around multiplication for low short product */
 #endif
 #elif (MULT == TOOM4)
@@ -75,7 +71,7 @@ init_list (unsigned int n)
   listz_t p;
   unsigned int i;
 
-  p = (mpz_t*) malloc (n * sizeof(mpz_t));
+  p = (mpz_t*) malloc (n * sizeof (mpz_t));
   if (p == NULL)
     return NULL;
   for (i = 0; i < n; i++)
@@ -83,6 +79,22 @@ init_list (unsigned int n)
   return p;
 }
 
+/* creates a list of n integers, return NULL if error. Allocates each
+   mpz_t to the size of n */
+listz_t
+init_list2 (unsigned int n, unsigned int N)
+{
+  listz_t p;
+  unsigned int i;
+
+  p = (mpz_t*) malloc (n * sizeof (mpz_t));
+  if (p == NULL)
+    return NULL;
+  for (i = 0; i < n; i++)
+    mpz_init2 (p[i], N);
+  return p;
+}
+
 /* clears a list of n integers */
 void
 clear_list (listz_t p, unsigned int n)
@@ -188,7 +200,7 @@ list_swap (listz_t p, listz_t q, unsigned int n)
 }
 
 /* p <- -q, keeps residues normalized */
-static void
+void
 list_neg (listz_t p, listz_t q, unsigned int l, mpz_t n)
 {
   unsigned int i;
@@ -246,13 +258,14 @@ list_mul_z (listz_t p, listz_t q, mpz_t r, unsigned int n, mpz_t m)
 }
 
 /* p <- gcd(n, l[0]*l[1]*...*l[k-1],
-   returns non-zero iff p is non trivial */
+   returns non-zero iff p is non trivial.
+   Clobbers l[0] */
 int
 list_gcd (mpz_t p, listz_t l, unsigned int k, mpz_t n)
 {
   unsigned int i;
   
-  for (i=1; i<k; i++)
+  for (i = 1; i < k; i++)
     {
       mpz_mul (l[0], l[0], l[i]);
       mpz_mod (l[0], l[0], n);
@@ -262,6 +275,22 @@ list_gcd (mpz_t p, listz_t l, unsigned int k, mpz_t n)
   return mpz_cmp_ui (p, 1);
 }
 
+
+/* Multiply up the integers in l, modulo n. Each entry becomes the
+   product (mod n) of itself and all previous entries */
+   
+void 
+list_mulup (mpz_t p, listz_t l, unsigned int k, mpz_t n, mpz_t t)
+{
+  unsigned int i;
+  
+  for (i = 1; i < k; i++)
+    {
+      mpz_mul (t, l[i - 1], l[i]);
+      mpz_mod (l[i], t, n);
+    }
+}
+
 /* p <- 0 */
 void
 list_zero (listz_t p, unsigned int n)
@@ -490,7 +519,7 @@ karatsuba (listz_t a, listz_t b, listz_t c, unsigned int K, listz_t t)
    The auxiliary array t contains at least list_mul_mem(l) entries.
    a and t should not overlap.
 */
-static void
+void
 list_mul (listz_t a, listz_t b, unsigned int k, int monic_b,
           listz_t c, unsigned int l, int monic_c, listz_t t)
 {
@@ -599,8 +628,8 @@ PolyFromRoots (listz_t G, listz_t a, unsigned int k, listz_t T, mpz_t n)
    if (k == 2)
      {
        mpz_mul (T[0], a[0], a[1]);
-       mpz_add (G[1], a[1], a[0]);
-       mpz_mod (G[1], G[1], n);
+       mpz_add (T[1], a[1], a[0]); /* mpz_add may allocate extra limb */
+       mpz_mod (G[1], T[1], n);
        mpz_mod (G[0], T[0], n);
        return;
      }
@@ -894,7 +923,9 @@ PrerevertDivision (listz_t a, listz_t b, listz_t invb,
       /* the high part of A * INVB is now in {t+K-2, K-1} */
       if (wrap)
 	{
-	  t2 = init_list (K - 1);
+	  MEMORY_TAG;
+	  t2 = init_list2 (K - 1, mpz_sizeinbase (n, 2));
+	  MEMORY_UNTAG;
 	  if (t2 == NULL)
 	    {
 	      fprintf (ECM_STDERR, "Error, not enough memory\n");
@@ -959,9 +990,9 @@ PrerevertDivision (listz_t a, listz_t b, listz_t invb,
    Returns 1 if a factor was found (stored in t), 0 otherwise.
 */
 int
-list_invert (listz_t inv, listz_t a, unsigned int l, mpz_t t, mpmod_t modulus)
+list_invert (listz_t inv, listz_t a, unsigned long l, mpz_t t, mpmod_t modulus)
 {
-  unsigned int i;
+  unsigned long i;
   
   if (l == 0)
     return 0;
