@@ -25,18 +25,8 @@ Proc. of ISSAC'03, Philadelphia, 2003.
 */
 
 #include <stdio.h>
-#include <stdlib.h>
-#include "gmp.h"
 #include "ecm-impl.h"
 
-#if !defined (_MSC_VER)
-#include <sys/time.h>
-#include <unistd.h>
-#if !defined (__MINGW32__)
-#include <sys/resource.h>
-#endif
-#endif
-
 #ifndef MAX
 #define MAX(a,b) (((a) > (b)) ? (a) : (b))
 #endif
@@ -385,7 +375,6 @@ TToomCookMul (listz_t b, unsigned int n,
     unsigned int nu, mu, h;
     unsigned int i;
     unsigned int btmp;
-    mpz_t comp_tmp;
     unsigned int tot_muls = 0;
 
     nu = n / 3 + 1;
@@ -463,8 +452,6 @@ TToomCookMul (listz_t b, unsigned int n,
     h = MAX(nu, mu);
     nu = mu = h;
 
-    mpz_init (comp_tmp);
-
     list_sub_safe (tmp, c + 3 * h, c + h,
                    (l + 1 > 3 * h ? l + 1 - 3 * h : 0), 
                    (l + 1 > h ? l + 1 - h : 0), 2 * h - 1);
@@ -525,8 +512,8 @@ TToomCookMul (listz_t b, unsigned int n,
     btmp = MIN(btmp, 2 * h - 1);
     for (i = 0; i < btmp; i++)
       {
-        mpz_mul_2exp (comp_tmp, c[h + i], 1);
-        mpz_add (tmp[5 * h - 2 + i], comp_tmp, tmp[3 * h - 1 + i]);
+        mpz_mul_2exp (tmp[5 * h - 2 + i], c[h + i], 1);
+        mpz_add (tmp[5 * h - 2 + i], tmp[5 * h - 2 + i], tmp[3 * h - 1 + i]);
       }
     while (i < 2 * h - 1)
       {
@@ -663,8 +650,6 @@ TToomCookMul (listz_t b, unsigned int n,
     list_add (b + 2 * h, b + 2 * h, tmp + h, n + 1 - 2 * h);
     /* b_{high} should be correct */
 
-    mpz_clear (comp_tmp);
-
     return tot_muls;
 }
 
