@@ -20,18 +20,19 @@
   MA 02111-1307, USA.
 */
 
-#include <stdio.h>
-#include "gmp.h"
 #include "ecm-impl.h"
 
-/* puts in {rp, n} the low part of {np, n} times {mp, n}
-   i.e. equivalent to:
+/* puts in {rp, n} the low part of {np, n} times {mp, n}, i.e. equivalent to:
+
+   mp_ptr tp;
+   TMP_DECL(marker);
+   TMP_MARK(marker);
    tp = TMP_ALLOC_LIMBS (2 * n);
    mpn_mul_n (tp, np, mp, n);
    MPN_COPY (rp, tp, n);
-   Remark: it is assumed at least 2*n limbs are allocated starting from rp.
+   TMP_FREE(marker);
  */
-static INLINE void
+void
 ecm_mul_lo_basecase (mp_ptr rp, mp_srcptr np, mp_srcptr mp, mp_size_t n)
 {
   mpn_mul_1 (rp, np, n, mp[0]);
@@ -39,10 +40,11 @@ ecm_mul_lo_basecase (mp_ptr rp, mp_srcptr np, mp_srcptr mp, mp_size_t n)
     mpn_addmul_1 (++rp, np, n, (++mp)[0]);
 }
 
-#define MPN_MUL_LO_THRESHOLD 32
-static mp_size_t threshold[MPN_MUL_LO_THRESHOLD] =
-/* {0,0,1,1,1,1,1,1,1,0,1,1,1,1,1,11,11,11,13,13,13,13,15,15,17,17,15,16,17,17,17,1}; */
-{0,0,0,0,0,1,1,0,0,1,1,1,10,9,1,1,12,1,1,1,1,15,16,14,18,19,17,18,19,20,21,22};
+#ifdef MPN_MUL_LO_THRESHOLD_TABLE
+size_t mpn_mul_lo_threshold[MPN_MUL_LO_THRESHOLD] = MPN_MUL_LO_THRESHOLD_TABLE;
+#else
+size_t mpn_mul_lo_threshold[MPN_MUL_LO_THRESHOLD];
+#endif
 
 
 void
@@ -52,7 +54,7 @@ ecm_mul_lo_n (mp_ptr rp, mp_srcptr np, mp_srcptr mp, mp_size_t n)
 
   if (n < MPN_MUL_LO_THRESHOLD)
     {
-      switch (k = threshold[n])
+      switch (k = mpn_mul_lo_threshold[n])
         {
         case 0:
           {
