@@ -20,15 +20,15 @@
   MA 02111-1307, USA.
 */
 
-#include <stdio.h>
 #include <stdlib.h>
-#if !defined (_MSC_VER)
-#include <unistd.h>
-#endif
-#include "gmp.h"
-#include "ecm.h"
+#include <string.h> /* for strlen */
 #include "ecm-impl.h"
 
+#if HAVE_UNISTD_H
+# include <unistd.h> /* for unlink */
+#endif
+
+
 #ifndef MAX
 #define MAX(a,b) (((a) > (b)) ? (a) : (b))
 #endif
@@ -138,7 +138,7 @@ print_vect (listz_t t, unsigned int l)
  * of the tree.
  */
 
-static void
+void
 TUpTree (listz_t b, listz_t *Tree, unsigned int k, listz_t tmp, int dolvl,
          unsigned int sh, mpz_t n, FILE *TreeFile)
 {
@@ -287,7 +287,9 @@ polyeval_tellegen (listz_t b, unsigned int k, listz_t *Tree, listz_t tmp,
       {
         outputf (OUTPUT_DEVVERBOSE, "polyeval_tellegen: allocating extra temp"
                  " space, want %d but T has only %d\n", tupspace, sizeT);
+        MEMORY_TAG;
         T = init_list (tupspace);
+        MEMORY_UNTAG;
 	if (T == NULL)
 	  return ECM_ERROR;
         allocated = 1;
@@ -322,17 +324,15 @@ polyeval_tellegen (listz_t b, unsigned int k, listz_t *Tree, listz_t tmp,
       {
         unsigned int lgk, i;
         FILE *TreeFile;
-        char fullname[256];
+        /* assume this "small" malloc will not fail in normal usage */
+	char *fullname = (char *) malloc (strlen (TreeFilename) + 1 + 2 + 1);
 
-        lgk = ceil_log2 (k);
+	lgk = ceil_log2 (k);
         for (i = 0; i < lgk; i++)
           {
-#ifdef HAVE_snprintf
-            snprintf (fullname, 256, "%.252s.%d", TreeFilename, i);
-#else
-            sprintf (fullname, "%.252s.%d", TreeFilename, i);
-#endif
-            TreeFile = fopen (fullname, "rb");
+            sprintf (fullname, "%s.%d", TreeFilename, i);
+            
+	    TreeFile = fopen (fullname, "rb");
             if (TreeFile == NULL)
               {
                 outputf (OUTPUT_ERROR, 
@@ -345,6 +345,7 @@ polyeval_tellegen (listz_t b, unsigned int k, listz_t *Tree, listz_t tmp,
             fclose (TreeFile);
             unlink (fullname);
           }
+        free (fullname);
       }
     else
       TUpTree (T, Tree, k, T + k, -1, 0, n, NULL);
