@@ -30,12 +30,17 @@
 
 #include <math.h>
 #include <stdlib.h>
-#include <stdio.h>
-#include <limits.h> /* for ULONG_MAX */
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-impl.h"
 
+#if HAVE_LIMITS_H
+# include <limits.h>
+#else
+# ifndef ULONG_MAX
+#  define ULONG_MAX __GMP_ULONG_MAX
+# endif
+#endif
+
+
 /******************************************************************************
 *                                                                             *
 *                                  Stage 1                                    *
@@ -54,19 +59,22 @@
 static void
 pp1_mul (mpres_t P1, mpres_t P0, mpz_t e, mpmod_t n, mpres_t P, mpres_t Q)
 {
-  unsigned long i;
   mp_size_t size_e;
+  unsigned long i;
+  int sign;
 
-  if (mpz_cmp_ui (e, 0) == 0)
+  sign = mpz_sgn (e);
+  mpz_abs (e, e);
+  if (sign == 0)
     {
       mpres_set_ui (P1, 2, n);
-      return;
+      goto unnegate;
     }
 
   if (mpz_cmp_ui (e, 1) == 0)
     {
       mpres_set (P1, P0, n);
-      return;
+      goto unnegate;
     }
   
   /* now e >= 2 */
@@ -101,8 +109,13 @@ pp1_mul (mpres_t P1, mpres_t P0, mpz_t e, mpmod_t n, mpres_t P, mpres_t Q)
         }
     }
 
-  mpz_add_ui (e, e, 1); /* recover original value of e */
   mpres_set (P1, P, n);
+  mpz_add_ui (e, e, 1); /* recover original value of e */
+unnegate:
+  if (sign == -1)
+    mpz_neg (e, e);
+  
+  return;
 }
 
 /* Input:  P0 is the initial point (sigma)
@@ -114,13 +127,14 @@ pp1_mul (mpres_t P1, mpres_t P0, mpz_t e, mpmod_t n, mpres_t P, mpres_t Q)
    Return value: non-zero iff a factor was found.
 */
 static int
-pp1_stage1 (mpz_t f, mpres_t P0, mpmod_t n, double B1, double B1done, mpz_t go)
+pp1_stage1 (mpz_t f, mpres_t P0, mpmod_t n, double B1, double *B1done, 
+            mpz_t go, int (*stop_asap)(void))
 {
   double B0, p, q, r;
   mpz_t g;
   mpres_t P, Q;
   mpres_t R, S, T;
-  int youpi;
+  int youpi = ECM_NO_FACTOR_FOUND;
   unsigned int max_size, size_n;
 
   mpz_init (g);
@@ -161,12 +175,18 @@ pp1_stage1 (mpz_t f, mpres_t P0, mpmod_t n, double B1, double B1done, mpz_t go)
   for (p = 2.0; p <= B0; p = getprime (p))
     {
       for (q = 1, r = p; r <= B1; r *= p)
-        if (r > B1done) q *= p;
+        if (r > *B1done) q *= p;
       mpz_mul_d (g, g, q, Q);
       if (mpz_sizeinbase (g, 2) >= max_size)
 	{
 	  pp1_mul (P0, P0, g, n, P, Q);
 	  mpz_set_ui (g, 1);
+          if (stop_asap != NULL && (*stop_asap) ())
+            {
+              outputf (OUTPUT_NORMAL, "Interrupted at prime %.0f\n", p);
+              *B1done = p;
+              goto clear_and_exit;
+            }
 	}
     }
 
@@ -175,22 +195,32 @@ pp1_stage1 (mpz_t f, mpres_t P0, mpmod_t n, double B1, double B1done, mpz_t go)
   /* then all primes > sqrt(B1) and taken with exponent 1 */
   for (; p <= B1; p = getprime (p))
     {
-      if (p > B1done)
+      if (p > *B1done)
         pp1_mul_prac (P0, (unsigned long) p, n, P, Q, R, S, T);
+  
+      if (stop_asap != NULL && (*stop_asap) ())
+        {
+          outputf (OUTPUT_NORMAL, "Interrupted at prime %.0f\n", p);
+          *B1done = p;
+          goto clear_and_exit;
+        }
     }
 
-  getprime (FREE_PRIME_TABLE); /* free the prime tables, and reinitialize */
+  *B1done = B1;
+  mpres_sub_ui (P, P0, 2, n);
+#ifndef FULL_REDUCTION
+  mpres_normalize (P); /* needed for gcd */
+#endif
+  mpres_gcd (f, P, n);
+  youpi = mpz_cmp_ui (f, 1);
 
+clear_and_exit:
+  getprime (FREE_PRIME_TABLE); /* free the prime tables, and reinitialize */
   mpres_clear (Q, n);
   mpres_clear (R, n);
   mpres_clear (S, n);
   mpres_clear (T, n);
   mpz_clear (g);
-
-  mpres_sub_ui (P, P0, 2, n);
-  mpres_gcd (f, P, n);
-  youpi = mpz_cmp_ui (f, 1);
-
   mpres_clear (P, n);
   
   return youpi;
@@ -308,12 +338,12 @@ addWnm (point *X, mpres_t P, mpmod_t modulus, unsigned int m,
    Return non-zero iff a factor was found (always zero in fact).
 */
 int
-pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF, 
-            mpres_t *x, listz_t t, int S, mpmod_t modulus)
+pp1_rootsF (listz_t F, root_params_t *root_params, unsigned long dF, 
+            mpres_t *x, listz_t t, mpmod_t modulus)
 {
-  unsigned int i, j;
+  unsigned long i, j;
   unsigned long muls = 0;
-  unsigned int st, st2;
+  long st, st1;
   int youpi = ECM_NO_FACTOR_FOUND;
   mpres_t fd[3];
   mpres_t u, v; /* auxiliary variables */
@@ -323,21 +353,21 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
   if (dF == 0)
     return youpi;
 
-  st2 = st = cputime ();
+  st1 = st = cputime ();
 
-  outputf (OUTPUT_DEVVERBOSE, "pp1_rootsF: d1 = %d, d2 = %d, dF = %d\n",
-	   d1, d2, dF);
+  outputf (OUTPUT_DEVVERBOSE, "pp1_rootsF: d1 = %lu, d2 = %lu, dF = %lu\n",
+	   root_params->d1, root_params->d2, dF);
 
   mpres_init (u, modulus);
   mpres_init (v, modulus);
 
-  if (ABS(S) == 1) /* special code with d1/6 muls */
+  if (ABS(root_params->S) == 1) /* special code with d1/6 muls */
     {
       mpres_init (fd[0], modulus);
       mpres_init (fd[1], modulus);
       mpres_init (fd[2], modulus);
   
-      mpz_set_ui (*t, d2);
+      mpz_set_ui (*t, root_params->d2);
       pp1_mul (fd[2], *x, *t, modulus, u, v);
       mpres_get_z (F[0], fd[2], modulus);
   
@@ -350,14 +380,14 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
       /* fd[0] = V_{7*d2}(P), fd[1] = V_{6*d2}(P), fd[2] = V_{d2}(P) */
 
       outputf (OUTPUT_VERBOSE,
-	       "Initializing table of differences for F took %ums\n",
-	       elltime (st2, cputime ()));
+	       "Initializing table of differences for F took %ldms\n",
+	       elltime (st1, cputime ()));
 
       i = 1;
       j = 7;
       while (i < dF)
 	{
-	  if (gcd (j, d1) == 1) /* (d2,d1) == 1 ==> (j*d2,d1) == (j,d1) */
+	  if (gcd (j, root_params->d1) == 1) /* (d2,d1) == 1 ==> (j*d2,d1) == (j,d1) */
 	    mpres_get_z (F[i++], fd[0], modulus);
 
 	  /* V_{m+n} = V_m * V_n - V_{m-n} */
@@ -377,10 +407,11 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
   else /* case |S| <> 1: this code works also for S=1, but is more
 	  expensive, since it can use up to 4*(d1/6) muls */
     {
-      init_roots_state (&state, S, d1, d2, 1.0);
+      init_roots_state (&state, root_params->S, root_params->d1, 
+                        root_params->d2, 1.0);
       mpz_set_ui (*t, 0);
-      coeffs = init_progression_coeffs (*t, state.dsieve, d2, 1, 6, state.S,
-                                        state.dickson_a);
+      coeffs = init_progression_coeffs (*t, state.dsieve, root_params->d2, 1, 
+                                        6, state.S, state.dickson_a);
       
       if (coeffs == NULL)
         return ECM_ERROR;
@@ -407,8 +438,8 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
       clear_list (coeffs, state.size_fd);
 
       outputf (OUTPUT_VERBOSE,
-	       "Initializing table of differences for F took %ums\n",
-	       elltime (st2, cputime ()));
+	       "Initializing table of differences for F took %ldms\n",
+	       elltime (st1, cputime ()));
 
       /* Now for the actual calculation of the roots. */
       for (i = 0; i < dF && !youpi;)
@@ -425,7 +456,7 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
 		}
 
 	      /* Is this a j value where we want Dickson(j*d2)*X as a root? */
-	      if (gcd (state.rsieve, d1) == 1)
+	      if (gcd (state.rsieve, root_params->d1) == 1)
 		{
 		  /* we have alpha^k = x * alpha + y
 		     thus alpha^k + beta^k = x * P + 2 * y.
@@ -455,7 +486,7 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
   mpres_clear (u, modulus);
   mpres_clear (v, modulus);
 
-  outputf (OUTPUT_VERBOSE, "Computing roots of F took %ums",
+  outputf (OUTPUT_VERBOSE, "Computing roots of F took %ldms",
 	   elltime (st, cputime ()));
   outputf (OUTPUT_DEVVERBOSE, " and %d muls", muls);
   outputf (OUTPUT_VERBOSE, "\n");
@@ -465,49 +496,47 @@ pp1_rootsF (listz_t F, unsigned int d1, unsigned int d2, unsigned int dF,
 
 /* return NULL if an error occurred */
 pp1_roots_state *
-pp1_rootsG_init (mpres_t *x, mpz_t s, unsigned int d1, unsigned int d2, 
-                 int S, mpmod_t modulus)
+pp1_rootsG_init (mpres_t *x, root_params_t *root_params, mpmod_t modulus)
 {
-  unsigned int st;
   mpres_t P;
-  mpz_t t;
   pp1_roots_state *state;
   unsigned long i;
 
-  ASSERT (gcd (d1, d2) == 1);
-  ASSERT (mpz_fdiv_ui (s, d1) == 0);
-
-  st = cputime ();
+  ASSERT (gcd (root_params->d1, root_params->d2) == 1);
 
   state = (pp1_roots_state *) malloc (sizeof (pp1_roots_state));
   if (state == NULL)
     return NULL;
 
-  state->S = ABS(S); /* we don't need the sign anymore after pp1_rootsG_init */
+  /* we don't need the sign anymore after pp1_rootsG_init */
+  state->S = ABS(root_params->S); 
 
   if (state->S == 1)
     {
+      mpz_t t;
       mpz_init (t);
       mpres_init (P, modulus);
       for (i = 0; i < 4; i++)
         mpres_init (state->tmp[i], modulus);
 
-      state->d = d1; /* needed in pp1_rootsG */
-      state->dsieve = d2; /* needed in pp1_rootsG */
-      /* We want to skip values where gcd(s + i * d1, d2) != 1 */
-      /* state->rsieve = s % d2 */
-      state->rsieve = mpz_fdiv_ui (s, d2);
+      state->dsieve = root_params->d2; /* needed in pp1_rootsG */
+      /* We want to skip values where gcd((i0 + i) * d1, d2) != 1.
+	 We can test for gcd(i0 + i, d2) instead and let pp1_rootsG()
+	 advance state->rsieve in steps of 1 */
+      /* state->rsieve = i0 % d2 */
+      state->rsieve = mpz_fdiv_ui (root_params->i0, root_params->d2);
       
-      outputf (OUTPUT_DEVVERBOSE, "pp1_rootsG_init: s = %Zd, state: d = %d, "
-               "dsieve = %d, rsieve = %d, S = %d\n", 
-               s, state->d, state->dsieve, state->rsieve, state->S);
+      outputf (OUTPUT_DEVVERBOSE, "pp1_rootsG_init: i0 = %Zd, state: "
+               "dsieve = %d, rsieve = %d, S = %d\n", root_params->i0, 
+               state->dsieve, state->rsieve, state->S);
       
-      pp1_mul (state->tmp[0], *x, s, modulus, state->tmp[3], P);
-      mpz_set_ui (t, d1);
+      mpz_set_ui (t, root_params->d1);
       pp1_mul (state->tmp[1], *x, t, modulus, state->tmp[3], P);
-      mpz_sub_ui (t, s, d1);
+      pp1_mul (state->tmp[0], state->tmp[1], root_params->i0, modulus, 
+               state->tmp[3], P);
+      mpz_sub_ui (t, root_params->i0, 1);
       mpz_abs (t, t);
-      pp1_mul (state->tmp[2], *x, t, modulus, state->tmp[3], P);
+      pp1_mul (state->tmp[2], state->tmp[1], t, modulus, state->tmp[3], P);
       /* for P+1, tmp[0] = V_s(P), tmp[1] = V_d1(P), tmp[2] = V_{|s-d1|}(P) */
 
       mpres_clear (P, modulus);
@@ -516,9 +545,9 @@ pp1_rootsG_init (mpres_t *x, mpz_t s, unsigned int d1, unsigned int d2,
   else
     {
       listz_t coeffs;
-      int dickson_a = (S < 0) ? -1 : 0;
-
-      state->nr = (d2 > 1) ? d2 - 1 : 1;
+      
+      state->dickson_a = (root_params->S < 0) ? -1 : 0;
+      state->nr = (root_params->d2 > 1) ? root_params->d2 - 1 : 1;
       state->size_fd = state->nr * (state->S + 1);
       state->next = 0;
       state->dsieve = 1;
@@ -531,7 +560,8 @@ pp1_rootsG_init (mpres_t *x, mpz_t s, unsigned int d1, unsigned int d2,
           return NULL;
         }
       
-      coeffs = init_progression_coeffs (s, d2, d1, 1, 1, state->S, dickson_a);
+      coeffs = init_progression_coeffs (root_params->i0, root_params->d2, 
+                 root_params->d1, 1, 1, state->S, state->dickson_a);
       if (coeffs == NULL)
         {
           free (state->fd);
@@ -584,12 +614,12 @@ pp1_rootsG_clear (pp1_roots_state *state, ATTRIBUTE_UNUSED mpmod_t modulus)
 }
 
 int
-pp1_rootsG (listz_t G, unsigned int dF, pp1_roots_state *state, mpmod_t modulus,
+pp1_rootsG (listz_t G, unsigned long dF, pp1_roots_state *state, mpmod_t modulus,
             mpres_t *x)
 {
-  unsigned int i;
+  unsigned long i;
   unsigned long muls = 0;
-  unsigned int st;
+  long st;
 
   st = cputime ();
 
@@ -615,7 +645,7 @@ pp1_rootsG (listz_t G, unsigned int dF, pp1_roots_state *state, mpmod_t modulus,
           mpres_swap (state->tmp[0], state->tmp[2], modulus);
           mpres_mul (state->tmp[3], state->tmp[2], state->tmp[1], modulus);
           mpres_sub (state->tmp[0], state->tmp[3], state->tmp[0], modulus);
-          state->rsieve = (state->rsieve + state->d) % state->dsieve;
+          state->rsieve++;
         }
     }
   else
@@ -652,9 +682,9 @@ pp1_rootsG (listz_t G, unsigned int dF, pp1_roots_state *state, mpmod_t modulus,
       mpres_clear (v, modulus);
     }
 
-  outputf (OUTPUT_VERBOSE, "Computing roots of G took %ums",
+  outputf (OUTPUT_VERBOSE, "Computing roots of G took %ldms",
 	   elltime (st, cputime ()));
-  outputf (OUTPUT_DEVVERBOSE, ", %u muls", dF);
+  outputf (OUTPUT_DEVVERBOSE, ", %lu muls", dF);
   outputf (OUTPUT_VERBOSE, "\n");
   
   return ECM_NO_FACTOR_FOUND;
@@ -677,15 +707,20 @@ pp1_rootsG (listz_t G, unsigned int dF, pp1_roots_state *state, mpmod_t modulus,
    Return value: non-zero iff a factor is found (1 for stage 1, 2 for stage 2)
 */
 int
-pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
-     mpz_t B2min_parm, mpz_t B2_parm, double B2scale, unsigned int k, int S,
-     int verbose, int repr, FILE *os, FILE *es, char *TreeFilename)
+pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double *B1done, double B1,
+     mpz_t B2min_parm, mpz_t B2_parm, double B2scale, unsigned long k, 
+     const int S, int verbose, int repr, int use_ntt, FILE *os, FILE *es, 
+     char *TreeFilename, double maxmem, gmp_randstate_t rng, 
+     int (*stop_asap)(void))
 {
-  int youpi = 0;
-  unsigned int st;
+  int youpi = ECM_NO_FACTOR_FOUND;
+  int po2 = 0;    /* Whether we should use power-of-2 poly degree */
+  long st;
   mpres_t a;
   mpmod_t modulus;
   mpz_t B2min, B2; /* Local B2, B2min to avoid changing caller's values */
+  unsigned long dF;
+  root_params_t root_params;
 
   set_verbose (verbose);
   ECM_STDOUT = (os == NULL) ? stdout : os;
@@ -701,29 +736,25 @@ pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
   st = cputime ();
 
   if (mpz_cmp_ui (p, 0) == 0)
-    {
-      gmp_randstate_t state;
-      gmp_randinit_default (state);
-      pm1_random_seed (p, n, state);
-      gmp_randclear (state);
-    }
+    pm1_random_seed (p, n, rng);
 
   mpz_init_set (B2min, B2min_parm);
   mpz_init_set (B2, B2_parm);
+  mpz_init (root_params.i0);
 
   /* Set default B2. See ecm.c for comments */
   if (ECM_IS_DEFAULT_B2(B2))
-    mpz_set_d (B2, pow (B2scale * 2.0 * B1 / 6.0, 1.424828748));
+    mpz_set_d (B2, B2scale * pow (B1 * PP1_COST, DEFAULT_B2_EXPONENT));
 
   /* set B2min */
   if (mpz_sgn (B2min) < 0)
     mpz_set_d (B2min, B1);
 
-  if (repr == 1)
+  if (repr == ECM_MOD_MPZ)
     mpmod_init_MPZ (modulus, n);
-  else   if (repr == MOD_MODMULN)
+  else   if (repr == ECM_MOD_MODMULN)
     mpmod_init_MODMULN (modulus, n);
-  else if (repr == MOD_REDC)
+  else if (repr == ECM_MOD_REDC)
     mpmod_init_REDC (modulus, n);
   else if (abs (repr) > 16)
     {
@@ -733,20 +764,31 @@ pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
   else /* automatic choice */
     mpmod_init (modulus, n, repr);
 
-  /* Set default degree for Brent-Suyama extension */
-  if (S == ECM_DEFAULT_S)
+  if (use_ntt)
+    po2 = 1;
+  
+  if (bestD (&root_params, &k, &dF, B2min, B2, po2, use_ntt, maxmem,
+             (TreeFilename != NULL), modulus) == ECM_ERROR)
     {
-      if (modulus->repr == MOD_BASE2 && modulus->Fermat > 0)
+      youpi = ECM_ERROR;
+      goto clear_and_exit;
+    }
+  
+  /* Set default degree for Brent-Suyama extension */
+  root_params.S = S;
+  if (root_params.S == ECM_DEFAULT_S)
+    {
+      if (modulus->repr == ECM_MOD_BASE2 && modulus->Fermat > 0)
         {
           /* For Fermat numbers, default is 1 (no Brent-Suyama) */
-          S = 1;
+          root_params.S = 1;
         }
       else
         {
           mpz_t t;
           mpz_init (t);
           mpz_sub (t, B2, B2min);
-          S = choose_S (t);
+          root_params.S = choose_S (t);
           mpz_clear (t);
         }
     }
@@ -754,25 +796,29 @@ pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
   if (test_verbose (OUTPUT_NORMAL))
     {
       outputf (OUTPUT_NORMAL, "Using ");
-      if (ECM_IS_DEFAULT_B1_DONE(B1done))
+      if (ECM_IS_DEFAULT_B1_DONE(*B1done))
         outputf (OUTPUT_NORMAL, "B1=%1.0f", B1);
       else
-        outputf (OUTPUT_NORMAL, "B1=%1.0f-%1.0f", B1done, B1);
+        outputf (OUTPUT_NORMAL, "B1=%1.0f-%1.0f", *B1done, B1);
       if (mpz_cmp_d (B2min, B1) == 0)
         outputf (OUTPUT_NORMAL, ", B2=%Zd", B2);
       else
         outputf (OUTPUT_NORMAL, ", B2=%Zd-%Zd", B2min, B2);
 
-      if (S > 0)
-        outputf (OUTPUT_NORMAL, ", polynomial x^%u", S);
+      if (root_params.S > 0)
+        outputf (OUTPUT_NORMAL, ", polynomial x^%u", root_params.S);
       else
-        outputf (OUTPUT_NORMAL, ", polynomial Dickson(%u)", -S);
+        outputf (OUTPUT_NORMAL, ", polynomial Dickson(%u)", 
+		 -root_params.S);
        /* don't print x0 in resume case */
-      if (ECM_IS_DEFAULT_B1_DONE(B1done)) 
+      if (ECM_IS_DEFAULT_B1_DONE(*B1done)) 
         outputf (OUTPUT_NORMAL, ", x0=%Zd", p);
       outputf (OUTPUT_NORMAL, "\n");
     }
 
+  outputf (OUTPUT_VERBOSE, "dF=%lu, k=%lu, d=%lu, d2=%lu, i0=%Zd\n", 
+           dF, k, root_params.d1, root_params.d2, root_params.i0);
+
   mpres_init (a, modulus);
   mpres_set_z (a, p, modulus);
 
@@ -783,15 +829,13 @@ pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
       outputf (OUTPUT_ERROR, "Error, maximal step1 bound for P+1 is %lu\n", 
                ULONG_MAX);
       youpi = ECM_ERROR;
-      goto clear_pp1;
+      goto clear_and_exit;
     }
 
-  if (B1 > B1done)
-    youpi = pp1_stage1 (f, a, modulus, B1, B1done, go);
+  if (B1 > *B1done)
+    youpi = pp1_stage1 (f, a, modulus, B1, B1done, go, stop_asap);
 
-  st = elltime (st, cputime ());
-
-  outputf (OUTPUT_NORMAL, "Step 1 took %ums\n", st);
+  outputf (OUTPUT_NORMAL, "Step 1 took %ldms\n", elltime (st, cputime ()));
   if (test_verbose (OUTPUT_RESVERBOSE))
     {
       mpz_t t;
@@ -802,18 +846,22 @@ pp1 (mpz_t f, mpz_t p, mpz_t n, mpz_t go, double B1done, double B1,
       mpz_clear (t);
     }
 
-  if (youpi == ECM_NO_FACTOR_FOUND) /* no factor found, no error */
-    youpi = stage2 (f, &a, modulus, B2min, B2, k, S, ECM_PP1, st, 
-                    TreeFilename);
+  mpres_get_z (p, a, modulus);
+
+  if (stop_asap != NULL && (*stop_asap) ())
+    goto clear_and_exit;
+      
+  if (youpi == ECM_NO_FACTOR_FOUND && mpz_cmp (B2, B2min) >= 0)
+    youpi = stage2 (f, &a, modulus, dF, k, &root_params, ECM_PP1, 
+                    use_ntt, TreeFilename, stop_asap);
 
   if (youpi > 0 && test_verbose (OUTPUT_NORMAL))
     pp1_check_factor (p, f); /* tell user if factor was found by P-1 */
 
-  mpres_get_z (p, a, modulus);
-
- clear_pp1:
+ clear_and_exit:
   mpres_clear (a, modulus);
   mpmod_clear (modulus);
+  mpz_clear (root_params.i0);
   mpz_clear (B2);
   mpz_clear (B2min);
 
